#include "nfsp_api.h"

#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <assert.h>
#include <netinet/in.h>

#include "config.h"
#include "default.h"
#include "logging.h"

// X functions in Xstat.c

// --------------- LOCKING -------------
// Thread safe
#if defined(_REENTRANT) || defined(__USE_REENTRANT) || defined(_THREAD_SAFE)
#  define TS 1
#  warning "THREAD SUPPORT ENABLED"
#endif

#ifdef TS
#  include <pthread.h>

// mutex on : configfile
// mutex on nfsp_iods.h variables: iods_nb iods
static pthread_mutex_t nfsp_api_mtx = PTHREAD_MUTEX_INITIALIZER;
#endif

static inline void lock_lib() {
#ifdef TS
	int ret;
	ret = pthread_mutex_lock(&nfsp_api_mtx);
	if (ret) {
		Dprintf(L_FATAL, "nfsp_init(): cannot acquire mutex");
	}
#endif
}

static inline void unlock_lib() {
#ifdef TS
	int ret;
	ret = pthread_mutex_unlock(&nfsp_api_mtx);
	if (ret) {
		Dprintf(L_FATAL, "nfsp_init(): cannot release mutex");
	}
#endif
}


// Note: accesses to theses variable must be lock'ed
#include "nfsp_iods.h"
static char *configfile = NULL;


/***************************************************************************/
int nfsp_init(const char * const iodsfile) {
	int ret;

	if (!iodsfile) {
		errno = EINVAL;
		return -1;
	}

	lock_lib();
	if (configfile) {
		Dprintf(D_LIBSTRIPING, "nfsp_init(): reconfiguring iod file");
		free(configfile);
	}
	configfile = strdup(iodsfile);

	// Parse config file
	init_iods(configfile);

	ret = (configfile) ? 0 : -1;
	unlock_lib();

	return ret;
}


/***************************************************************************/
int nfsp_get_iods(struct nfsp_iod_t **ppiods, int *psize) {
	struct nfsp_iod_t *ptr;
	int i;
	
	if (!ppiods || !psize) {
		errno = EINVAL;
		return -1;
	}

	lock_lib();
	ptr = (struct nfsp_iod_t *) calloc(iods_nb, sizeof(struct nfsp_iod_t));
	if (!ptr) {
		return -1;
	}

	for (i = 0; i < iods_nb; i++) {
		ptr[i].ip = iods[i].sin_addr.s_addr;
		ptr[i].port = iods[i].sin_port;
		ptr[i].path = NULL;		// FIXME
	}

	*ppiods = ptr;
	*psize = iods_nb;
	unlock_lib();

	return 0;
}

void nfsp_free_iods(struct nfsp_iod_t *piods) {
	if (piods) {
		return;
	}

	free(piods);
}

/***************************************************************************/

int nfsp_get_localfn_from_xstat(char **plocalfn, const struct Xstat * const pxs) {
	const char *keyformat = "%2.2x/i%8.8xs%8.8x";
	char buf[2 + 1 + 1 + 8 + 1 + 8 + 1];
	unsigned int hash;
	unsigned char c;
	char *localfn;
	
	if (!plocalfn || !pxs) {
		errno = EINVAL;
		return -1;
	}

	hash = (unsigned int) pxs->mfstat.st_ino ^ (unsigned int) pxs->nstat.seed;
        hash ^= hash >> 16;
	c = (0xFF & hash) % NB_IODS_SUBDIR;

	sprintf(buf, keyformat, c, pxs->mfstat.st_ino, pxs->nstat.seed);
	
	localfn = strdup(buf);
	if (!localfn) {
		plocalfn = NULL;
		return -1;
	}


	*plocalfn = localfn;
	return 0;
}

void nfsp_free_localfn(char * localfn) {
	if (!localfn) {
		return;
	}
	free(localfn);
}

/***************************************************************************/

/* ppiov	OUT	allocates the output array
 * piov_size	OUT	the array size
 * nbiods	IN	the number of iods
 * pxs		IN	Xstat information
 * offset	IN	offset to which access is started
 * size		IN	length of accessed block
 *
 * return values
 *  0		success
 *  !0		failure (errno is set)
 *
 * */
int nfsp_get_iov_from_xstat_r(struct nfsp_iov_t **ppiov, int *piov_size,
		 const int nbiods,
		 const struct Xstat * const pxs,
	       	 const off_t offset, const size_t size) {

	off_t o;	
	size_t toread;
	struct nfsp_iov_t *iov;
	int i, iov_size;
	char *localfn;

	if (!ppiov || !piov_size || !pxs) {
		errno = EINVAL;
		return -1;
	}

	if (size <= 0) {
		*ppiov = NULL;
		*piov_size = 0;
		return 0;
	}
	toread = size;
#if 0
	// To handle read and write we do not consider the file size
	if (offset + size > pxs->nstat.realsize) {
		toread = pxs->nstat.realsize - offset;
	} else {
		toread = size;
	}
#endif



	//from o to o + toread - 1
	iov_size = ((offset + toread - 1) / BLOCKSIZE) - (offset / BLOCKSIZE) + 1;
	Dprintf(D_LIBSTRIPING, "nfsp_get_iov: offset=%lld toread=%d", (int64_t) offset, toread);
	Dprintf(D_LIBSTRIPING, "nfsp_get_iov: iov_size=%d", iov_size);
	iov = (struct nfsp_iov_t *) calloc(iov_size, sizeof(struct nfsp_iov_t));
	if (!iov) {
		return -1;
	}

	if (nfsp_get_localfn_from_xstat(&localfn, pxs)) {
		return -1;
	}

	Dprintf(D_LIBSTRIPING, "nfsp_get_iov: localfn[%s]", localfn);

	o = offset;
	i = 0;
	while (toread > 0) {
		size_t chunksize;
		
		chunksize = BLOCKSIZE - (o % BLOCKSIZE);
		if (chunksize > toread) {
			chunksize = toread;
		}

		// See nfs_dispatch.c:find_iod()
		iov[i].iodno = (pxs->nstat.seed + (o / BLOCKSIZE)) % nbiods; 

		
		// See iodng.c::key_2fninternal()
		iov[i].localfn = localfn;

		iov[i].wanted_offset = o;
		
		// See iodng.c::do_reply_read()
		iov[i].mapped_offset = ((o / (nbiods * BLOCKSIZE)) * BLOCKSIZE) + (o % BLOCKSIZE);

		iov[i].size = chunksize;

		o = (o / BLOCKSIZE) * BLOCKSIZE + BLOCKSIZE;
		toread -= chunksize;
		i++;
	}

	*ppiov = iov;
	*piov_size = iov_size;
	
	return 0;
}

int nfsp_get_iov_from_xstat(struct nfsp_iov_t **ppiov, int *piov_size,
		 const struct Xstat * const pxs,
	       	 const off_t offset, const size_t size) {
	int ret;

	lock_lib();
	ret = nfsp_get_iov_from_xstat_r(ppiov, piov_size, iods_nb, pxs, offset, size);
	unlock_lib();

	return ret;
}


/* ppiov	OUT	allocates the output array
 * piov_size	OUT	the array size
 * filein	IN	file to Xlstat()
 * offset	IN	offset to which access is started
 * size		IN	length of accessed block
 *
 * return values
 *  0		success
 *  !0		failure (errno is set)
 *
 * */
int nfsp_get_iov_from_filename(struct nfsp_iov_t **ppiov, int *piov_size,
		 const char * const filein, const off_t offset, const size_t size) {
	
	struct Xstat xs;

	if (!filein) {
		errno = EINVAL;
		return -1;
	}
	if (-1 == access(filein, F_OK)) {
		*ppiov = NULL;
		*piov_size = 0;
		return -1;
	}

	if (-1 == Xlstat(filein, &xs)) {
		return -1;
	}

	return nfsp_get_iov_from_xstat(ppiov, piov_size, &xs, offset, size);
}



/***************************************************************************/
void nfsp_free_iov(struct nfsp_iov_t *piov) {
	if (!piov) {
		return;
	}
	nfsp_free_localfn(piov[0].localfn);
	free(piov);
}

/*******************************************************************************/
/* Parameters must be set in xs
 * - If filein does not exist and xs->nstat is filled (size to 0), it is created 
 * 
 * Note: mfstat handling should be done
 * */
int nfsp_meta_update(const char * const filein, const struct Xstat * const xs) {
	int create;		// Is it a file creation ? (boolean)
	int ret;
	int fd;
	struct nstat ns;

	// A few checks
	if (!filein || !xs) {
		errno = EINVAL;
		return -1;
	}

	create = (-1 == access(filein, F_OK));
	if (create) {
		if (xs->nstat.magic != XSTAT_MAGIC) {
			Dprintf(L_WARNING, "not enough magic");
			return -1;
		}
		if (xs->nstat.seed == 0) {
			Dprintf(D_LIBSTRIPING, "don't forget to set the seed");
		}
		if (xs->nstat.realsize != 0) {
			Dprintf(L_WARNING, "nstat.realsize must be set at 0 for file creation (is %lld)", (int64_t) xs->nstat.realsize);
			return -1;
		}

		fd = creat(filein, xs->mfstat.st_mode);
		if (fd == -1) {
			Dprintf(L_WARNING, "creat failed errno = %d", errno);
			return -1;
		}


		if (-1 == fchown(fd, xs->mfstat.st_uid, xs->mfstat.st_gid)) {
			Dprintf(D_LIBSTRIPING, "fchown failed errno = %d", errno);
			//return -1;
		}
	
		/* xs->mfstat.st_atime xs->mfstat.st_mtime xs->mfstat.st_ctime
		 */

		memcpy(&ns, &(xs->nstat), sizeof(ns));

	} else {
		fd = open(filein, O_RDWR);
		if (fd == -1) {
			Dprintf(L_WARNING, "cannot open errno=%d", errno);
			return -1;
		}

		ret = pread(fd, &ns, sizeof(ns), 0);
		if (ret != sizeof(ns)) {
			if (ret == -1) {
				Dprintf(L_WARNING, "pread failed errno=%d", errno);
			} else {
				Dprintf(L_WARNING, "short read %dB instead of %dB", ret, sizeof(ns));
			}
			close(fd);
			return -1;
		}

		if (ns.magic != XSTAT_MAGIC) {
			Dprintf(L_WARNING, "not enough magic");
			close(fd);
			return -1;
		}

                if (xs->nstat.realsize < ns.realsize) {
			// truncating file
			ns.realsize = xs->nstat.realsize;
                } else {
			// extending file
			ns.realsize = xs->nstat.realsize;
		}

		if (-1 == fchown(fd, xs->mfstat.st_uid, xs->mfstat.st_gid)) {
			Dprintf(D_LIBSTRIPING, "fchown failed errno = %d", errno);
			//return -1;
		}
	}

	ret = pwrite(fd, &ns, sizeof(ns), 0);
	if (ret != sizeof(ns)) {
		if (ret == -1) {
			Dprintf(L_WARNING, "pwrite failed errno=%d", errno);
		} else {
			Dprintf(L_WARNING, "short pwrite %dB instead of %dB", ret, sizeof(ns));
		}
		Dprintf(L_WARNING, "CRITICAL: realsize should be %lld seed is=%d", (int64_t) ns.realsize, ns.seed);
	}

	if (-1 == close(fd)) {
		Dprintf(L_WARNING, "close failed errno=%d", errno);
		return -1;
	}

	return 0;
}


// Note: the mutex is not taken as there are not many problem but losing some
// logs which is acceptable since syslog may lose some too
void nfsp_dbglog_enable() {
	enable_logging("libstriping");
}
void nfsp_dbglog_disable() {
	enable_logging("general");
}

// TODO fix these issues with foreground/background/stderr/...
void nfsp_log_background() {
	background_logging();
}

