/* NFSP striping API
 *
 * Author:
 *  2002-2003 Pierre Lombard <pierre.lombard@imag.fr>
 *
 * License: GPLv2 or LGPLv2
 *
 * Changelog:
 * V4 added (un)lock() support when compiled with _REENTRANT
 *    added nfsp_get_iov_from_xstat_r()
 *    added nfsp_dbglog_enable() nfsp_dbglog_disable()
 * V3 added nfsp_get_localfn_from_xstat() and nfsp_free_localfn()
 * V2 added nfsp_get_iov_from_xstat() 
 * V2 nfsp_get_iov() renamed to nfsp_get_iov_from_filename()
 * */

#ifndef __NFSP_API_H__
#define __NFSP_API_H__ 1

#define NFSP_API_VERSION 4

#include <sys/types.h>
#include <errno.h>
#include <inttypes.h>

// Xstat prototypes
// extended stat/fstat/lstat (to stat metafiles)
#include "Xstat.h"

struct nfsp_iod_t {
	uint32_t ip;		// Network ordered IP
	uint16_t port;		// Network ordered port
	char *path;		// Currently always NULL
};

struct nfsp_iov_t {
	int iodno;		// iod no
	char *localfn;		// local filename (i.e. on the iod)
	off_t wanted_offset;	// logical offset
	off_t mapped_offset;	// physical offset in localfn
	size_t size;		// request size
};

// Initialize the library
//
// Note:
//   - always call this first
//   - needs the library lock
int nfsp_init(const char * const iodsfile);

// Gets an array of IOD's
//
// ppiods	OUT	array of iods (is allocated)
// psize	OUT	size of the array
// 
// Return values:
//   -1		error
//   0		success
//
// Note:
//   - needs the library lock
int nfsp_get_iods(struct nfsp_iod_t **ppiods, int *psize);

// Free an array of IOD's
// piods	IN	array of iods (already allocated)
void nfsp_free_iods(struct nfsp_iod_t *piods);

// The two functions below nfsp_get_iov_from_* let the user get an access
// vector (to read or write).  Such a vector indicates where and on which range
// I/O may be done thanks to plain read() or write().
//
// ppiov	OUT	access vector (is allocated)
// piov_size	OUT	vector size
// nbiods	IN	the number of iods
// pxs		IN	the xstat information of the metafile
// filein	IN	the metafile name
// offset	IN	start offset
// size		IN	size to access
//
// return values:
//    0		success
//   !0		error    (errno is set)

// thread safe - no initialization needed
int nfsp_get_iov_from_xstat_r(struct nfsp_iov_t **ppiov, int *piov_size,
		 const int nbiods,
		 const struct Xstat * const pxs,
	       	 const off_t offset, const size_t size);

// Note:
//   - needs the library lock
int nfsp_get_iov_from_xstat(struct nfsp_iov_t **ppiov, int *piov_size,
		 const struct Xstat * const pxs,
	       	 const off_t offset, const size_t size);

// Note:
//   - needs the library lock
int nfsp_get_iov_from_filename(struct nfsp_iov_t **ppiov, int *piov_size,
		 const char * const filein, const off_t offset, const size_t size);

// Free an allocated nfsp_iov_t*
// piov		IN	access vector (already allocated)
void nfsp_free_iov(struct nfsp_iov_t *piov);


// Return and allocate the filename on the iod (local filename) 
// based on the Xstat structure given
// plocalfn	OUT	filename (allocated - free with nfsp_free_localfn)
// pxs		IN	xstat structure
//
// return values:
//    0		success
//   !0		error    (errno is set)
int nfsp_get_localfn_from_xstat(char **plocalfn, const struct Xstat * const pxs);

// Free an allocated local filename string
// localfn	IN/OUT
void nfsp_free_localfn(char * localfn);


// Update/Create a metafile's property (size, etc, ...)
//
// Note: if filein does not exist and if xs.nstat.size is set to 0, then 
//       filein is created (empty)
// 
// filein	IN	filename (metafile)
// xs		IN	extended struct stat
//
// NOTE: - there is a small window between the access() and the creat() call
//         proper fix TODO
//       - except that, this function should be reentrant 
int nfsp_meta_update(const char * const filein, const struct Xstat * const xs);


// The default is to only echo the important logs (warnings, fatal errors) You
// cannot disable important logs but they may be sent to the syslog by means of
// nfsp_log_background()

// Enable full debug log 
void nfsp_dbglog_enable();

// Disable full debug log
void nfsp_dbglog_disable();

#endif
