// xstat (extended stat) functions
// functions to handle metafiles
// 
// Basic metafile handling
// The metadata is stored in two places : 
//   1. the underlying fs (acces via stat() call on the metafile 
//      most stats come from here (uid, permissions, etc . . .)
//   2. the content of the file, which is the binary dump of a
//      struc nstat (see in Xstat.h)
//
// Xstat only works with REGULAR files (i.e. plain data file)
// 
// Author: Pierre Lombard <pierre.lombard@imag.fr>
// Licence: GPL 

#include "Xstat.h"

#include <fcntl.h>
#include <stdio.h>
#include <errno.h>

#define ERR(x...) do { fprintf(stderr, ## x); } while (0)

// Fills buf with the metafile opened in fd (at least O_RDONLY)
// fd  IN   must be readable
// buf OUT  always overwritten
int Xfstat(int fd, struct Xstat *buf) {
	int ret;

	ret = fstat(fd, &(buf->mfstat));
	if (ret == -1) {
		ERR("cannot fstat errno=%d", errno);
		return -1;
	}

	if (!S_ISREG(buf->mfstat.st_mode)) {
		// Not a regular file that's not an error but
		// we don't have to handle this
		return 0;
	}
	
	// Using a pread instead of a simple read avoids doing annoying things
	// with lseek ...
	ret = pread(fd, (char* )(&(buf->nstat)), sizeof(struct nstat), 0);
	if (ret != sizeof(struct nstat)) {
		if (ret == -1) {
			ERR("read error errno=%d", errno);
			return -1;
		} else {
			ERR("FIXME metafile has an incorrect size=%d instead of %d",
					ret, sizeof(struct nstat));
			// XXX What to do ?
			return -1;
		}
	}

	if (buf->nstat.magic != XSTAT_MAGIC) {
		ERR("metafile has a bad magic number");
		return -1;
	}

	return 0;
}


// This function is used for stat and lstat calls
typedef int (*stat_proto)(const char *, struct stat *);
static inline int do_foostat(stat_proto f, const char *file_name, struct Xstat *buf) {
	int ret, fd;

	struct stat *mfs = &buf->mfstat;	// MetaFile STAT's: concerns the metafile
	struct nstat *ns = &buf->nstat;	// New STAT's: concerns the real file

	ret = f(file_name, mfs);

	if (-1 == ret) {
		return -1;
	}

	if (!S_ISREG(mfs->st_mode)) {
		// S_ISDIR(m)  directory?
		// S_ISLNK(m)  symbolic link? (Not in POSIX.1-1996.)
		//   We can't do much more for a directory or a symlink: it's not a 
		//   regular file so the stat we have will be enough
		// S_ISCHR(m)  character device?
		// S_ISBLK(m)  block device?
		// S_ISFIFO(m) fifo?
		// S_ISSOCK(m) socket? (Not in POSIX.1-1996.)
		return ret;
	}

	fd = open(file_name, O_RDONLY);
	if (-1 == fd) {
		return -1;
	}
	
	// Using a pread instead of a simple read avoids doing annoying things
	// with lseek ...
	ret = pread(fd, (char* )ns, sizeof(struct nstat), 0);
	if (ret != sizeof(struct nstat)) {
		if (ret == -1) {
			ERR("read error errno=%d", errno);
			close(fd);
			return -1;
		} else {
			ERR("FIXME metafile has an incorrect size=%d instead of %d", ret, sizeof(struct nstat));
			// XXX What to do ?
			close(fd);
			return -1;
		}
	}

	if (ns->magic != XSTAT_MAGIC) {
		ERR("metafile has a bad magic number");
		close(fd);
		return -1;
	}

	ret = close(fd);
	if (ret == -1) {
		ERR("close failed errno=%d", errno);
		return -1;
	}

	return 0;
}

// fills buf with the metafile named file_name
int Xstat(const char *file_name, struct Xstat *buf) {
	return do_foostat(stat, file_name, buf);
}

// fills buf with the metafile named file_name
int Xlstat(const char *file_name, struct Xstat *buf) {
	return do_foostat(lstat, file_name, buf);
}

